#define UNICODE
#pragma warning(disable: 4786)

#include <stdio.h>
#include <windows.h>
#include <string>
#include <map>
#include <vector>
#include <crtdbg.h>
#define QWORD ULONGLONG
#include "LrfParser.h"

bool g_DumpBinary=false;
bool g_DumpXml=false;

void LRFHeader1::Dump(FILE* out)
{
  fwprintf(out, L"<Header1 ");
  fwprintf(out, L" version = \"%d\"", version);
  fwprintf(out, L" xorKey = \"0x%02X\"", xorKey);
  fwprintf(out, L" RootObject = \"0x%04X\"", rootObjectId);
  fwprintf(out, L" ObjectsCount = \"%Ld\"\n", nObjects);
  fwprintf(out, L" ObjectsTable = \"0x%08LX\"", offObjectTable);
  fwprintf(out, L"/>\n");
}

void LRFHeader2::Dump(FILE* out)
{
  fwprintf(out, L"<Header2 ");
  fwprintf(out, L" d1_20 = \"0x%08X\"", d1_20);
  fwprintf(out, L" b1_24 = \"%d\"", b1_24);
  fwprintf(out, L" w1_26 = \"%d\"", w1_26);
  fwprintf(out, L" w2_2A = \"%d\"", w2_2A);
  fwprintf(out, L" w3_2C = \"%d\"", w3_2C);
  fwprintf(out, L" b2_2E = \"%d\"", b2_2E);
  fwprintf(out, L" signature = \"[0x%08X, 0x%08X, 0x%08X, 0x%08X, 0x%08X]\"", signature[0], signature[1], signature[2], signature[3], signature[4]);
  fwprintf(out, L" dwTocID = \"0x%08X\"", dwTocID);
  fwprintf(out, L" dwTocOffset = \"0x%08X\"", dwTocOffset);
  fwprintf(out, L" XMLSize = \"0x%04X\"", XMLSize);
  fwprintf(out, L" w5_4E = \"0x%04X\"", w5_4E);
  fwprintf(out, L" GifSize = \"0x%08X\"", GifSize);
  fwprintf(out, L"/>\n");
}

bool LRFObject::ReadTags()
{
  Clear();
  DWORD left = m_dataSize;
  LRFTag* curTag=0;
  if (!m_pFile->Seek(m_offset))
    return false;
  while (left>0)
  {
    QWORD curPos = m_pFile->Pos();
    curTag = new LRFTag(m_pFile);
    if (!curTag->Load())
    {
      delete curTag;
      fwprintf(stderr, L"Encountered bad tag; stop parsing object (offset 0x%08I64X, 0x%04X bytes left)\n", curPos, left);
      return true;
    }
    else
    {
      switch (curTag->id())
      {
        case 0xF500: //object type
          m_ObjectType = *(WORD*)(curTag->data()+4);
          m_objectName = m_ObjectType<0x1f?ObjectTypes[m_ObjectType]:L"<invalid>";
          wchar_t objID[8];
          swprintf(objID, L"_%04X", m_id);
          m_objectName += objID;
          break;
        case 0xF504: //stream size
          m_StreamDataSize = *(DWORD*)(curTag->data());
          break;
        case 0xF554: //stream flags
          m_StreamFlags = *(WORD*)(curTag->data());
          break;
        case 0xF505: //stream start
          m_StreamDataOffset = m_pFile->Pos();
          m_pFile->Seek(m_StreamDataOffset+m_StreamDataSize);
          left-=m_StreamDataSize;
          break;
      }
      left -= curTag->dataSize();
      left -= sizeof(WORD);
      m_tags.push_back(curTag);
    }
  }
  return true;
}

bool LRFObject::DumpTOC(FILE* out, BYTE* buf, DWORD bufLen)
{
  WORD w,c;
  BYTE *p;
  DWORD d;
  c = *(WORD*)buf;
  p = buf+4*(c+1);
  while (c)
  {
    fwprintf(out, L"<TOCEntry>", d);
    d = *(DWORD*)p;
    if (m_pFile->isValidObject(d))
      fwprintf(out, L"%s,", m_pFile->getObjectName(d)->c_str());
    else
      fwprintf(out, L"0x%04X,", d);
    p+=4;
    d = *(DWORD*)p;
    if (m_pFile->isValidObject(d))
      fwprintf(out, L"%s,", m_pFile->getObjectName(d)->c_str());
    else
      fwprintf(out, L"0x%04X,", d);
    p+=4;
    w = *(WORD*)p;
    p+=2;
    fwprintf(out, L"%.*s", w/2, p);
    fwprintf(out, L"</TOCEntry>\n", d);
    p+=w;
    c--;
  }
  return true;
}

bool LRFObject::DumpStreamTags(FILE* out, BYTE* buf, DWORD bufLen)
{
  WORD w;
  BYTE *p;
  p = buf;
  MemReader mr;
  LRFTag tag(&mr);
  while (bufLen)
  {
    w = *(WORD*)p;
    TagInfo& info = StdTags[w&0xFF];
    if ( (w>>8!=0xF5) || (info.paramSize == -1))
    {
      //not a tag; just print it out
      fwrite(p, 1, 2, out);
      bufLen -= 2;
      p+=2;
    }
    else
    {
      mr.SetBuf(p, bufLen);
      if (!tag.Load())
      {
        fwprintf(stderr, L"Bad tag: 0x%04X\n", w);
        return false;
      }
      tag.Dump(out, m_pFile);
      DWORD taglen = mr.Pos();
      bufLen -= taglen;
      p+=taglen;
    }
  }
  return true;
}

bool LRFObject::DumpStream(FILE* out, const std::wstring& filename)
{
  if (!g_DumpBinary && !g_DumpXml) 
    return true; //nothing to do

  if (!m_pFile->Seek(m_StreamDataOffset))
  {
    fwprintf(stderr, L"Fatal: can't seek to offset 0x%08X\n. Not writing object stream.\n",m_StreamDataOffset);
    return false;
  }

  Reader* src = m_pFile;
  DWORD bufSize = m_StreamDataSize;
  if (m_StreamFlags & 0x100)
  {
    if (!src->Read(&bufSize, 4))
    {
      fwprintf(stderr, L"Fatal: can't read from file. Not writing object stream.\n",m_StreamDataOffset);
      return false;
    }
    src = new Unpacker(src, m_StreamDataSize-4);
    //fwprintf(stderr, L"Don't know how to unpack streams yet.\n");
    //return true;
  }
  if (m_StreamFlags & 0x200)
  {
    DWORD len = bufSize;
    if (len > 0x400 && (m_ObjectType == 0x11 || m_ObjectType == 0x17 || m_ObjectType == 0x19))
      len = 0x400;
    WORD xorKey = (m_StreamDataSize % m_pFile->m_pHeader1->xorKey) + 0x0F;
    src = new Descrambler(src, xorKey, len);//!!! LEAK
  }

  if (m_StreamFlags & 0x8000)
  {
    fwprintf(stderr, L"Don't know how to decrypt streams yet.\n");
    //return true;
  }

  BYTE* buf = new BYTE[bufSize];
  if (!buf)
  {
    fwprintf(stderr, L"Out of memory while dumping a stream. Couldn't allocate %d bytes.\n",m_StreamDataSize);
    return false;
  }
  if (!src->Read(buf, bufSize))
  {
    fwprintf(stderr, L"Fatal: error reading object stream.\n");
    delete buf;
    return false;
  }

  /*if (m_StreamFlags & 0x200)
  {
    DWORD xorKey = (m_StreamDataSize % m_pFile->m_pHeader1->xorKey) + 0x0F;
    xorKey &= 0xFF;
    xorKey = xorKey<<8  | xorKey; 
    xorKey = xorKey<<16 | xorKey; //propagate low byte to all four

    DWORD len = m_StreamDataSize;
    if (len > 0x400 && (m_ObjectType == 0x11 || m_ObjectType == 0x17 || m_ObjectType == 0x19))
      len = 0x400;
    DWORD* p1=(DWORD*)buf;
    while (len>4)
    {
      *p1^=xorKey;
      p1++,len-=4;
    }
    if (len) *p1^=xorKey;
  }*/
  if (g_DumpBinary)
  {
    FILE* dump = _wfopen((filename+L".stream").c_str() ,L"wb");
    if (!dump)
    {
      fwprintf(stderr, L"Fatal: error creating file %s.\n", filename.c_str());
      delete buf;
      return false;
    }
    if (bufSize!=fwrite(buf, 1, bufSize, dump))
    {
      fwprintf(stderr, L"Fatal: error writing to file %s.\n", filename.c_str());
      delete buf;
      return false;
    }
    fclose(dump);
  }
  if (g_DumpXml)
  {
    switch (m_ObjectType)
    {
      case 0x03: //Header
      case 0x04: //Footer
      case 0x06: //Block
      case 0x0D: //canvas
      case 0x02: //page
      case 0x0a: //Text
      case 0x1d: //SimpleText
      {
        FILE* dumpXml = _wfopen((filename+L".xml").c_str() ,L"wb");
        fprintf(dumpXml,"\xFF\xFE");//BOM
        fwprintf(dumpXml, L"<?xml version=\"1.0\"?>\n<Stream>\n");
        DumpStreamTags(dumpXml, buf, bufSize);
        fwprintf(dumpXml, L"</Stream>\n");
        fclose(dumpXml);
        break;
      }
      case 0x1e: //TOC
      {
        FILE* dumpXml = _wfopen((filename+L".xml").c_str() ,L"wb");
        fprintf(dumpXml,"\xFF\xFE");//BOM
        fwprintf(dumpXml, L"<?xml version=\"1.0\"?>\n<TOC>\n");
        DumpTOC(dumpXml, buf, bufSize);
        fwprintf(dumpXml, L"</TOC>\n");
        fclose(dumpXml);
        break;
      }
    }
  }
  delete buf;
  return true;
}

void LRFObject::Dump(FILE* out)
{
  fwprintf(out, L"<Object ID=\"0x%04X\"", m_id);
  fwprintf(out, L" offset = \"0x%08X\"", m_offset);
  fwprintf(out, L" size   = \"0x%08X\"", m_dataSize);
  fwprintf(out, L" name   = \"%s\"", m_objectName.c_str());
  fwprintf(out, L">\n");
  for (std::vector<LRFTag*>::iterator i=m_tags.begin(); i!=m_tags.end(); i++)
  {
    (*i)->Dump(out, m_pFile);
    if ((*i)->id()==0xF505 && m_StreamDataOffset) //StreamStart
    {
      fwprintf(out, L"<StreamData offset =\"0x%04X\" size = \"0x%04X\">\n", m_StreamDataOffset, m_StreamDataSize);
      DumpStream(out, m_objectName);
      fwprintf(out, L"</StreamData>\n");
    }
  }
  fwprintf(out, L"</Object>\n");
}
void LRFObject::Clear()
{
   for (std::vector<LRFTag*>::iterator i=m_tags.begin(); i!=m_tags.end(); i++)
     delete (*i);
}

BYTE* LRFTag::ReserveBuffer(DWORD size)
{
  m_dataSize = size;
  if (m_dataSize<=sizeof(m_shortData))
  {
    return m_shortData;
  }
  else
  {
    m_customData = new BYTE[m_dataSize];
    if (!m_customData)
      return NULL;
    return m_customData;
  }
}

bool LRFTag::Load()
{
   if (!m_pFile->Read(&m_tagID, sizeof(m_tagID)))
     return false;
   TagInfo& info = StdTags[m_tagID&0xFF];
   if ( (m_tagID>>8!=0xF5) || (info.paramSize == -1))
   {
     //fwprintf(stderr, L"Bad tag: 0x%04X at 0x%08I64X\n", m_tagID, file->Pos());
     return false; //invalid tag
   }
   if (info.pCustomParser)
     return info.pCustomParser(m_pFile, this);
   BYTE* buf = ReserveBuffer(info.paramSize);
   if (!buf)
     return false;
   else
     return m_pFile->Read(buf, info.paramSize);
}

void LRFTag::Dump(FILE* out, LRFFile* pLRFFile)
{
  fwprintf(out, L"<Tag ID=\"0x%04X\"", m_tagID);
  if (g_TagNames.find(m_tagID)!=g_TagNames.end())
    fwprintf(out, L" name = \"%s\"", g_TagNames[m_tagID].c_str());
  fwprintf(out, L" size = \"0x%08X\"", m_dataSize);
  fwprintf(out, L">\n");
  BYTE* p=data();
  /*if (m_tagID==0xF500)
  {
    WORD objType=*(WORD*)(p+4);
    fwprintf(out, L"<TagObjectType type=\"%s\"/>", objType<0x1f?ObjectTypes[objType]:L"<invalid>");
  }*/
  WORD w;
  DWORD d;
  size_t i;
  if (m_dataSize)
    switch (m_tagID)
    {
      case 0xF516:
      case 0xF555:
      case 0xF559:
      case 0xF55A:
      case 0xF55D: //a string
        fwprintf(out, L"\"%s\"\n", p+2);
        break;
      case 0xF578: //empdotcode
        d = *(DWORD*)p;
        w = *(WORD*)(p+4);
        fwprintf(out, L"0x%08X, 0x%04X, \"%s\"\n", d, w, p+8);
        break;
      case 0xF5F9: //???
        w = *(WORD*)p;
        d = *(DWORD*)(p+2);
        fwprintf(out, L"0x%04X, ", w);
        if (pLRFFile->isValidObject(d))
          fwprintf(out, L"%s\n", pLRFFile->getObjectName(d)->c_str());
        else
          fwprintf(out, L"0x%04X\n", d);
        break;
      case 0xF54A: //image rect
        d = *(DWORD*)p;
        fwprintf(out, L"(%d, %d, ", d&0xffff, d>>16);
        d = *(DWORD*)(p+4);
        fwprintf(out, L"%d, %d)\n", d&0xffff, d>>16);
        break;
      case 0xF549: //image on canvas
        d = *(DWORD*)p;
        fwprintf(out, L"%d, %d, ", d&0xffff, d>>16);
        d = *(DWORD*)(p+4);
        if (pLRFFile->isValidObject(d))
          fwprintf(out, L"%s\n", pLRFFile->getObjectName(d)->c_str());
        else
          fwprintf(out, L"0x%04X\n", d);
        break;
      case 0xF54B: //image dimensions
        d = *(DWORD*)p;
        fwprintf(out, L"%dx%d\n", d&0xffff, d>>16);
        break;
      case 0xF50B:
      case 0xF55C:
      case 0xF56C: //a list
        if (m_tagID==0xF56C)
          w = 2;
        else
        {
          w = *(WORD*)p;
          p+=2;
        }
        fwprintf(out, L"[ ");
        for (i=0; i<w; i++)
        {
          if (i>0)
            fwprintf(out, L", ");
          d = *(DWORD*)p;
          if (pLRFFile->isValidObject(d))
            fwprintf(out, L"%s", pLRFFile->getObjectName(d)->c_str());
          else
            fwprintf(out, L"0x%04X", d);
          p+=4;
        }
        fwprintf(out, L" ]\n");
        break;
      default:
        if (m_dataSize == 4)
        {
          d = *(DWORD*)p;
          switch (m_tagID)
          {
            case 0xF504:
            case 0xF534:
            case 0xF537:
            case 0xF517:
            case 0xF518:
            case 0xF5F2:
              fwprintf(out, L"0x%04X\n", d);
              break;
            default:
              if (pLRFFile->isValidObject(d))
                fwprintf(out, L"%s\n", pLRFFile->getObjectName(d)->c_str());
              else
                fwprintf(out, L"0x%04X\n", d);
          }
        }
        else
        {
          if (m_dataSize == 2)
          {
            w = *(WORD*)p;
            switch (m_tagID)
            {
              case 0xF554:
                fwprintf(out, L"0x%04X\n", w);
                break;
              default:
                fwprintf(out, L"%hd\n", w);
            }
          }
          else
          {
            for (size_t i=0; i<m_dataSize; i++)
              fwprintf(out, L"%02X ", p[i]);
            fwprintf(out, L"\n");
          }
        }
    }
  fwprintf(out, L"</Tag>\n");
}

bool LRFFile::Open()
{
   m_pFile = _wfopen(m_filename.c_str(),L"rb");
   if (!m_pFile)
   {
     m_LastError=L"Can't open input file: "+m_filename+L'.';
     return false;
   }
   m_pHeader1 = new LRFHeader1;
   if (!Read(m_pHeader1, sizeof(LRFHeader1)))
   {
     m_LastError=L"Error reading header 1 in "+m_filename+L'.';
     return false;
   }

   if (((DWORD*)m_pHeader1)[1] == 0x58524C ) //'LRX'
   {
     //printf("LRX file. Seeking to LRF header\n");
     m_offsetInLRX = 0;
     Seek(0x14);
     Read(&m_offsetInLRX, 4);
     Seek(0);
     if (!Read(m_pHeader1, sizeof(LRFHeader1)))
     {
       m_LastError=L"Error reading header 1 in "+m_filename+L'.';
       return false;
     }
   }

   if (memcmp(&m_pHeader1->signature, &LRFSignature, 8))
   {
     m_LastError=L"Not an LRF file: "+m_filename+L'.';
     return false;
   }

   if (!Seek(H2OFF))
   {
     m_LastError=L"Error seeking input file: "+m_filename+L'.';
     return false;
   }

   m_pHeader2 = new LRFHeader2;
   if (!Read(m_pHeader2, (m_pHeader1->version>=800)?H2SIZE:H2SIZEOLD))
   {
     m_LastError=L"Error reading header 2 in "+m_filename+L'.';
     return false;
   }

   if (!ReadObjects())
     return false;
   return true;
}

void LRFFile::DumpMetaInfo(const std::wstring& filename)
{  
  //seek to compressed metainfo
  fwprintf(stderr,L"seeking to 0x%04X\n", H2OFF+((m_pHeader1->version>=800)?H2SIZE:H2SIZEOLD));
  if (!Seek(H2OFF+((m_pHeader1->version>=800)?H2SIZE:H2SIZEOLD)))
  {
    fwprintf(stderr,L"Error seeking input file: %s.\n", m_filename.c_str());
    return;
  }
  DWORD uncompSize;
  if (!Read(&uncompSize, 4))
  {
    fwprintf(stderr,L"Error reading input file: %s.\n", m_filename.c_str());
    return;
  }
  fwprintf(stderr,L"uncompSize=0x%04X\n", uncompSize);
  Unpacker u(this, m_pHeader2->XMLSize);
  BYTE* buf = new BYTE[uncompSize];
  if (!u.Read(buf, uncompSize))
  {
    fwprintf(stderr,L"Error decompressing metainfo from: %s.\n", m_filename.c_str());
    delete buf;
    return;
  }
  FILE* dumpXml = _wfopen(filename.c_str() ,L"wb");
  if (!dumpXml)
  {
    fwprintf(stderr,L"Error creating %s\n",filename.c_str());
    delete buf;
    return;
  }
  if (uncompSize!=fwrite(buf,1,uncompSize,dumpXml))
  {
    fwprintf(stderr,L"Error writing to %s\n",filename.c_str());
    delete buf;
    return;
  }
  fclose(dumpXml);
  delete buf;
}
void LRFFile::Dump(FILE *out)
{
   fwprintf(out, L"<?xml version=\"1.0\"?>\n<LRF>\n");
   m_pHeader1->Dump(out);
   m_pHeader2->Dump(out);
   DumpObjects(out);
   DumpMetaInfo(L"metainfo.xml");
   fwprintf(out, L"</LRF>");

}

void LRFFile::DumpObjects(FILE* out)
{
  for (mapDWtoObj::iterator i = m_ObjectsMap.begin(); i != m_ObjectsMap.end(); i++)
  {
     (*i).second->Dump(out);
  }
}

bool LRFFile::ReadObjects()
{
   LRFObjectRecord* objectsTable = new LRFObjectRecord[m_pHeader1->nObjects];
   if (!objectsTable)
   {
     m_LastError=L"Not enough memory for objects table!";
     return false;
   }
   if (!Seek(m_pHeader1->offObjectTable))
   {
     m_LastError=L"Error seeking to the objects offsets table in "+m_filename+L'.';
     return false;
   }

   if (!Read(objectsTable, sizeof(LRFObjectRecord)*m_pHeader1->nObjects))
   {
     m_LastError=L"Error reading input file: "+m_filename+L'.';
     return false;
   }

   for (DWORD i=0; i<m_pHeader1->nObjects; i++)
   {
     LRFObjectRecord& curObject = objectsTable[i];
     LRFObject* obj = new LRFObject(curObject.id, curObject.offset, curObject.size, this);
     obj->ReadTags();
     m_ObjectsMap.insert(mapDWtoObj::value_type(curObject.id, obj));
   }
   delete objectsTable;
   return true;
}

void Usage()
{
  wprintf(L"BBeB parser 0.2 (c) 2004 roxfan\n");
  wprintf(L"Usage: LrfParser file.lrf [-d] [-x]\n");
  wprintf(L"       -d dump binary streams\n");
  wprintf(L"       -x dump known stream formats to xml\n");
}

void ReadTagnames(wchar_t* argv0)
{
  std::wstring tags = argv0;
  int slash = tags.find_last_of(L"\\/:");
  if (slash == std::wstring::npos)
    slash = 0;
  tags = tags.substr(0, slash) + L"\\tagnames.txt";
  FILE* fTagNames = _wfopen(tags.c_str(),L"r");
  if (!fTagNames)
    fwprintf(stderr,L"Couldn't open tagnames.txt, no tag names will be provided.\n");
  else
  {
    while (!feof(fTagNames))
    {
      WORD tagID;
      char name[1024];
      wchar_t wname[1024];
      if (2==fscanf(fTagNames, "%hX: %s\n", &tagID, name))
      {
        mbstowcs(wname, name, 1024);
        g_TagNames[tagID] = std::wstring(wname);
      }
    }
    fclose(fTagNames);
  }
}

int wmain(int argc, wchar_t* argv[])
{
  //_CrtSetBreakAlloc(53);
  if (argc<2)
  {
    Usage();
    return 1;
  }
  for (int i=1; i<argc; i++)
  {
    std::wstring arg(argv[i]);
    if (arg[0]==L'/' || arg[0]==L'-')
      switch(arg[1])
      {
        case L'?':
        case L'h':
        case L'H':
        {
          Usage();
          exit(1);
        }
        case L'd':
        case L'D':
          g_DumpBinary = true;
          break;
        case L'x':
        case L'X':
          g_DumpXml = true;
          break;
      }
    
  }
  ReadTagnames(argv[0]);
  //int f = _CrtSetDbgFlag(_CRTDBG_REPORT_FLAG);
  //_CrtSetDbgFlag(f|_CRTDBG_CHECK_ALWAYS_DF|_CRTDBG_LEAK_CHECK_DF);
  LRFFile lrf(argv[1]);
  if (!lrf.Open())
  {
    wprintf(L"Open error: %s\n", lrf.m_LastError.c_str());
    return 1;
  }
  FILE* dump=fopen("dump.xml","wb");
  fprintf(dump,"\xFF\xFE");//BOM
  lrf.Dump(dump);
  fclose(dump);
  return 0;
}